##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
from exputils.data.logger import Logger
from typing import Optional

# holds the global logger object
log = Logger()


def reset():
    """
    Resets the log which deletes all data in the memory and resets all changed configuration such
    as the directory path of the log.

    :warning: Data that has been logged after the [save][exputils.data.logging.save] function
    was called will be lost.
    """
    global log
    log = Logger()


def get_log():
    # returns the current logger.
    global log
    return log


def set_log(new_log):
    #Sets the given logger to be the global log
    global log
    log = new_log


def set_directory(directory: str):
    """
    Sets the directory path under which the logs will be saved.
    The default is `./data`.

    If the directory does not exist it will be created.

    Parameters:
        directory (str):
            Path to the directory.

    """
    log.directory = directory


def get_directory() -> str:
    """
    Returns the path to the directory the log.

    Returns:
        directory (str):
            Path to the directory.
    """
    return log.directory


def contains(name: str) -> bool:
    """
    Check if a log entry for the given name exists.

    Parameters:
        name (str): The name to be checked in the log.

    Returns:
        is_contained (bool): True if a log for the name exists, otherwise False.
    """
    return (name in log)


def clear(name: Optional[str] = None):
    """
    Clears the data of all or a specific log entry.

    :warning: Data that has been logged after the [save][exputils.data.logging.save] function
    was called will be lost.

    Parameters:
        name (str):
            Name of the log entry.
            If no name is given, then all log entries will be cleared.
    """
    log.clear(name=name)


def get_item(name: str) -> object:
    """
    Returns the logged data for a certain entry.

    Parameters:
        name (str):
            Name of the log entry.

    Returns:
        Logged data. Usually in form of a numpy array.
    """
    return log[name]


def add_value(name: str,
              value,
              log_to_tb: Optional[bool] = None,
              tb_global_step: Optional[int] = None,
              tb_walltime: Optional[float] = None):
    """
    Adds a value to a log entry with optional parallel TensorBoard logging.

    Parameters:
        name (str):
            The name of the entry where the value is added.
        value (Any):
            The value to be added. Can be a scalar or an array.
        log_to_tb (bool):
            Defines of the value should be logged to TensorBoard in parallel to the standard log.
            If True, log the value to TensorBoard.
            If False, do not log the value to TensorBoard.
            If not specified, then it gets logged if TensorBoard is globally activated.
            See [activate_tensorboard][exputils.data.logging.activate_tensorboard] for more details.
        tb_global_step (int):
            If logging to TensorBoard is active, then this is the global step value to record with
            the value in TensorBoard.
        tb_walltime (float):
            If logging to TensorBoard is active, then this is an optional override for the walltime
            in TensorBoard.
    """
    log.add_value(name, value, log_to_tb, tb_global_step, tb_walltime)


def add_scalar(name: str,
              scalar,
              log_to_tb: Optional[bool] = None,
              tb_global_step: Optional[int] = None,
              tb_walltime: Optional[float] = None):
    """
    Adds a scalar value to a log entry with optional parallel TensorBoard logging.

    Note: Has the same functionality as [add_value][exputils.data.logging.add_value] and exists to
    have a similar named log function as TensorBoard.

    Parameters:
        name (str):
            The name of the entry where the value is added.
        scalar (Any):
            The scalar value to be added.
        log_to_tb (bool):
            Defines of the value should be logged to TensorBoard in parallel to the standard log.
            If True, log the value to TensorBoard.
            If False, do not log the value to TensorBoard.
            If not specified, then it gets logged if TensorBoard is globally activated.
            See [activate_tensorboard][exputils.data.logging.activate_tensorboard] for more details.
        tb_global_step (int):
            If logging to TensorBoard is active, then this is the global step value to record with
            the value in TensorBoard.
        tb_walltime (float):
            If logging to TensorBoard is active, then this is an optional override for the walltime
            in TensorBoard.
    """
    log.add_scalar(name, scalar, log_to_tb, tb_global_step, tb_walltime)


def add_histogram(name: str,
                  values,
                  log_to_tb: Optional[bool] = None,
                  tb_global_step: Optional[int] = None,
                  tb_walltime: Optional[float] = None):
    """
    Adds a histogram which is a one-dimensional array a log entry with optional parallel TensorBoard
    logging.

    This allows to add the values as a histogram plot to TensorBoard.

    Parameters:
        name (str):
            The name of the entry where the value is added.
        values (Any):
            The array of values to be added.
        log_to_tb (bool):
            Defines of the value should be logged to TensorBoard in parallel to the standard log.
            If True, log the value to TensorBoard.
            If False, do not log the value to TensorBoard.
            If not specified, then it gets logged if TensorBoard is globally activated.
            See [activate_tensorboard][exputils.data.logging.activate_tensorboard] for more details.
        tb_global_step (int):
            If logging to TensorBoard is active, then this is the global step value to record with
            the value in TensorBoard.
        tb_walltime (float):
            If logging to TensorBoard is active, then this is an optional override for the walltime
            in TensorBoard.
    """
    log.add_histogram(name, values, log_to_tb, tb_global_step, tb_walltime)


def get_values(name: str):
    """
    Returns the logged data for a certain entry.

    Note: Has the same functionality as [get_item][exputils.data.logging.get_item] and exists to
    have a similar named log function as TensorBoard.

    Parameters:
        name (str):
            Name of the log entry.

    Returns:
        Logged data. Usually in form of a numpy array.
    """
    return log[name]


def add_object(name: str,
               obj: object):
    """
    Adds an object to a log entry. Objects are stored in a list and saved as dill files.

    Parameters:
        name (str):
            The name of the log entry where the object is added.
        obj (object):
            The object to be added to the log.
    """
    log.add_object(name, obj)


def get_objects(name: str) -> list:
    """
    Returns the logged objects for a certain entry.

    Args:
        name (str): Name of the log entry.

    Returns:
        objects (list): Logged objects.
    """
    return log[name]


def add_single_object(name: str,
                      obj: object,
                      directory: Optional[str] = None):
    """
    Logs a single object which is directly written to a dill file and not stored in memory.

    Parameters:
        name (str):
            The name of the object which is used for the filename.
        obj (object):
            The object to be logged.
        directory (str):
            Optional directory path where the dill file for the object is saved.
            Default is the log directory.
    """
    log.add_single_object(name, obj, directory=directory)


def items() -> list:
    """
    Returns all log entries as a list of tuples with the name and values of the entries.

    Returns:
        entries (list): All logged entries.
    """
    return log.items()


def save(directory: Optional[str] = None):
    """
    Saves the log.
    All logged values are stored in memory and only written to disk when this function is called.

    Parameters:
        directory (str):
            Optional directory path where the dill file for the object is saved.
            Default is the log directory.
    """
    log.save(directory=directory)


def load(directory: Optional[str] = None,
         load_objects: bool = False):
    """
    Loads entries from a log directory into the log.
    Afterwards the loaded entries can be accessed via the [items][exputils.data.logging.items] and
    [get_item][exputils.data.logging.get_item] functions.

    Parameters:
        directory (str):
            Optional directory path where the dill file for the object is saved.
            Default is the log directory.
        load_objects (bool):
            True if objects (dill files) that are in the directory should also be loaded.
            Default is False to avoid unintended large loads of objects.
    """
    log.load(directory=directory, load_objects=load_objects)


def load_single_object(name: str) -> object:
    """
    Loads a single object from the log folder and returns it.
    The object is not stored in the log memory.

    Parameters:
        name (str):
            Name of the object which is used for the filename.
            Either with or without the `'.dill'` extension.

    Returns:
        obj (object): Loaded object.
    """
    return log.load_single_object(name)


def set_config(config: Optional[dict] = None,
               **kwargs):
    # sets the config of the underlying Logger object
    # see logger.py for more information
    log.config = eu.combine_dicts(kwargs, config, log.config)


####################################################
# TENSORBOARD

def tensorboard():
    """
    Returns the tensorboard SummaryWriter object used to log values to TensorBoard.

    Returns:
        writer (SummaryWriter): TensorBoard SummaryWriter object.
    """
    return log.tensorboard


def create_tensorboard(config: Optional[dict] = None,
                       **kwargs):
    """
    Creates the SummaryWriter object used to log values to TensorBoard.
    This allows to set its configuration parameters.

    For more details see: https://pytorch.org/docs/stable/tensorboard.html

    Parameters:
        config (dict):
            Optional dictionary with the configuration of the SummaryWriter.
            Has the same entries as the set of "Other Parameters" below.

    Other Parameters:
        log_dir (str):
            Directly location of the TensorBoard log files.
            Default is `experiments/tensorboard_logs/exp_\<experiment_id>/rep_\<repetition_id>/\<date>_\<time>.`
        purge_step (int):
            When logging crashes at step T+XT+XT+X and restarts at step TTT, any events whose
            global_step larger or equal to TTT will be purged and hidden from TensorBoard.
            Note that crashed and resumed experiments should have the same log_dir.
        max_queue (int):
            Size of the queue for pending events and summaries before one of the ‘add’ calls forces
            a flush to disk. (default = 10)
        flush_secs (int):
            How often, in seconds, to flush the pending events and summaries to disk. (default = 120)
        filename_suffix (string):
            Suffix added to all event filenames in the log_dir directory. (default = '.tblog')

    Returns:
        writer (SummaryWriter): TensorBoard SummaryWriter object.
    """

    return log.create_tensorboard(config=config, **kwargs)


def activate_tensorboard(config: Optional[dict] = None,
                         **kwargs):
    """
    Activates parallel TensorBoard logging.
    When it is activated, then the logging functions (
    [add_value][exputils.data.logging.add_value],
    [add_scalar][exputils.data.logging.add_scalar],
    [add_histogram][exputils.data.logging.add_histogram])
    will automatically log each value also in TensorBoard if not otherwise specified for them.

    Creates a TensorBoard SummaryWriter if non is created yet with the [create_tensorboard][exputils.data.logging.create_tensorboard] function.
    If non is created then the configureation of the writer can be defined by the parameters of this function.
    For more details see: https://pytorch.org/docs/stable/tensorboard.html

    Other Parameters:
        log_dir (str):
            Directly location of the TensorBoard log files.
            Default is `experiments/tensorboard_logs/exp_\<experiment_id>/rep_\<repetition_id>/\<date>_\<time>.`
        purge_step (int):
            When logging crashes at step T+XT+XT+X and restarts at step TTT, any events whose
            global_step larger or equal to TTT will be purged and hidden from TensorBoard.
            Note that crashed and resumed experiments should have the same log_dir.
        max_queue (int):
            Size of the queue for pending events and summaries before one of the ‘add’ calls forces
            a flush to disk. (default = 10)
        flush_secs (int):
            How often, in seconds, to flush the pending events and summaries to disk. (default = 120)
        filename_suffix (string):
            Suffix added to all event filenames in the log_dir directory. (default = '.tblog')

    Returns:
        writer (SummaryWriter): TensorBoard SummaryWriter object.
    """

    return log.activate_tensorboard(config=config, **kwargs)


def deactivate_tensorboard():
    """
    Deactivates tensorboard logging.
    Afterwards, values will not be automatically logged via the add_value / add_scalar function
    to the tensorboard.
    """
    return log.deactivate_tensorboard()


def is_tensorboard_active() -> bool:
    """Returns true, if the tensorboard is active.

    Returns:
        is_active (bool): True if the tensorboard is active, otherwise False.
    """
    return log.is_tensorboard_active






##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.gui.jupyter.multi_selection_widget import MultiSelectionWidget
import exputils as eu
import numpy as np

class RepetitionIDsSelectionWidget(MultiSelectionWidget):
    """
    MultiSelectionWidget to select experiment ids in a experiment_data dict.
    """

    @staticmethod
    def default_config():
        dc = eu.gui.jupyter.MultiSelectionWidget.default_config()

        # if true, returns 'all' for selected_experiment_ids property if all repetitions
        # are selected instead of a list of all repetition ids
        dc.is_return_all_string = True

        # defines if an extra choice called none is given
        dc.is_select_non_choice = True

        return dc

    def __init__(self, experiment_data, experiment_descriptions=None, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())

        # get experiment ids
        self.repetition_ids = []

        # choice strings that are displayed to the user, can include ids, name, short_name, ...
        all_repetition_ids = []

        # only extract ids automatically, if an experiment description is given
        if experiment_descriptions is not None:
            for exp_descr in experiment_descriptions.values():
                if 'repetition_ids' in exp_descr:
                    all_repetition_ids += exp_descr['repetition_ids']

        repetition_ids = np.unique(all_repetition_ids).tolist()

        super().__init__(choices=repetition_ids, **self.config)


    @property
    def selected_repetition_ids(self):

        if self.config.is_return_all_string and self.config.is_select_all_choice and self.select_all_checkbox.value == True:
            return 'all'
        else:
            return self.selected_choices


    @selected_repetition_ids.setter
    def selected_repetition_ids(self, selected_repetition_ids):
        self.selected_choices = selected_repetition_ids


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.gui.jupyter.multi_selection_widget import MultiSelectionWidget
import warnings
import exputils as eu

class ExperimentIDsSelectionWidget(MultiSelectionWidget):
    '''
    MultiSelectionWidget to select experiment ids in a experiment_data dict.
    '''

    @staticmethod
    def default_config():
        dc = eu.gui.jupyter.MultiSelectionWidget.default_config()

        dc.title = 'TITLE_LABELS'
        dc.title_labels = []

        dc.label_template = '<name> <description>'

        dc.title_label_template = '<short_name>'

        # if true, returns 'all' for selected_experiment_ids property if all experiments
        # are selected instead of a list of all experiments
        dc.is_return_all_string = False

        return dc

    def __init__(self, experiment_data, experiment_descriptions=None, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())

        # get experiment ids
        self.experiment_ids = []

        # choice strings that are displayed to the user, can include ids, name, short_name, ...
        choices = []
        title_labels = []

        if experiment_descriptions is None:
            # detect ids from the experiment data
            for exp_id in experiment_data.keys():
                self.experiment_ids.append(exp_id)
                choices.append(exp_id)
                title_labels.append(exp_id)

        else:

            # show experiments in the order defined by the descriptions
            ordered_exp_ids = eu.data.get_ordered_experiment_ids_from_descriptions(experiment_descriptions)

            # detect ids from the experiment_descriptions
            for exp_id in ordered_exp_ids:
                exp_descr =  experiment_descriptions[exp_id]

                if exp_descr['is_load_data']:
                    self.experiment_ids.append(exp_descr['id'])

                    choice_str = eu.misc.replace_str_from_dict(
                        self.config.label_template,
                        exp_descr,
                        pattern_format='<{}>')
                    choices.append(choice_str)

                    title_str = eu.misc.replace_str_from_dict(
                        self.config.title_label_template,
                        exp_descr,
                        pattern_format='<{}>')
                    title_labels.append(title_str)

        if not self.config.title_labels:
            self.config.title_labels = title_labels

        super().__init__(choices, **self.config)

    @property
    def selected_experiment_ids(self):

        if self.config.is_return_all_string and self.is_all_selected:
            selected_choices = 'all'
        else:
            # find experiment_ids to selected choices
            selected_choices = []
            selected_choices_idxs = self.selected_choices_idxs
            for selected_idx in selected_choices_idxs:
                selected_choices.append(self.experiment_ids[selected_idx])

        return selected_choices

    @selected_experiment_ids.setter
    def selected_experiment_ids(self, selected_experiment_ids):

        if selected_experiment_ids == 'all':
            self.selected_choices = 'all'
        else:
            selected_choices_idxs = []
            for exp_id in selected_experiment_ids:
                # find exp_id in  self.experiment_ids
                if exp_id in self.experiment_ids:
                    selected_choices_idxs.append(self.experiment_ids.index(exp_id))
                else:
                    warnings.warn('Experiment id {!r} does not exists in list of experiment ID\'s. Setting will be ignored.'.format(exp_id))

            self.selected_choices_idxs = selected_choices_idxs


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from __future__ import annotations # see https://stackoverflow.com/questions/63460126/typeerror-type-object-is-not-subscriptable-in-a-function-signature
import os
import sys
import exputils as eu
from glob import glob
import re
import numpy as np
import warnings
import collections
import importlib.util
from exputils.misc.attrdict import AttrDict
from typing import Optional
from types import ModuleType


# TODO: Feature - allow to load data from several campaigns

def load_experiment_descriptions(experiments_directory: Optional[str] = None,
                                 allowed_experiments_id_list: Optional[list] = None,
                                 denied_experiments_id_list: Optional[list] = None,
                                 experiment_directory_template: Optional[str] = None,
                                 repetition_directory_template: Optional[str] = None) -> AttrDict:
    """
    Loads and returns descriptions of experiments from a specified experiments directory.

    Arguments:
        experiments_directory (str): Path to the experiments directory.
            Defaults to "..\DEFAULT_EXPERIMENTS_DIRECTORY".
        allowed_experiments_id_list (list): List of experiment IDs to be loaded.
            Cannot be used with denied_experiments_id_list.
            Default: All experiments are considered.
        denied_experiments_id_list (list): List of experiment IDs to be excluded.
            Cannot be used with allowed_experiments_id_list.
            Default: All experiments are considered.
        experiment_directory_template (str): Template string for the name of experiment directories.
            The template should include a placeholder for the experiment id.
            Example: 'experiment_{:06d}' for experiment folders with ids with at least six digits.
            Defaults to EXPERIMENT_DIRECTORY_TEMPLATE.
        repetition_directory_template (str): Template string for the name of experiment directories.
            The template should include a placeholder for the repetition id.
            Example: 'repetition_{:06d}' for repetition folders with ids with at least six digits.
            Defaults to REPETITION_DIRECTORY_TEMPLATE.

    Returns:
        AttrDict: A dictionary containing descriptions of the experiments.
        The keys are the experiment ids and the values are dictionaries with the follwing properties:

            - id: experiment id.
            - name: Name of the experiment. Default: 'exp <experiment id\>'.
            - short_name: A short name for display purposes. Default: 'e<experiment id\>'.
            - order: Sort order of experiments if they should be displayed in a GUI.
            - directory: Full path to experiment directory.
            - description: Description of the experiment. Default: ''.
            - repetition_ids: List of repetition ids.
            - repetition_directories : List of full paths to repetition directories.
    """

    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    if allowed_experiments_id_list is not None and denied_experiments_id_list is not None:
        raise ValueError('allowed_experiments_id_list and denied_experiments_id_list can not be set at the same time!')

    if experiment_directory_template is None: experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE
    experiment_directory_template = re.sub('\{.*\}', '*', experiment_directory_template)

    if repetition_directory_template is None: repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE
    repetition_directory_template = re.sub('\{.*\}', '*', repetition_directory_template)

    experiment_descriptions = AttrDict()

    exp_directories = glob(os.path.join(experiments_directory, experiment_directory_template))
    for order, exp_directory in enumerate(np.sort(exp_directories)):

        try:
            exp_id = re.findall(r'\d+', os.path.basename(exp_directory))[0]
        except IndexError as err:
            raise ValueError('The experiments_directory (\'{}\') seems not to have experiment folders!'.format(experiments_directory)) from err

        is_add_experiment_descr = True
        if allowed_experiments_id_list is not None and exp_id not in allowed_experiments_id_list:
            is_add_experiment_descr = False
        elif denied_experiments_id_list is not None and exp_id in denied_experiments_id_list:
            is_add_experiment_descr = False

        if is_add_experiment_descr:
            experiment_descr = AttrDict()
            experiment_descr.id = exp_id
            experiment_descr.name = 'exp {}'.format(exp_id)
            experiment_descr.order = order
            experiment_descr.is_load_data = True
            experiment_descr.directory = exp_directory
            experiment_descr.short_name = 'e{}'.format(exp_id)
            experiment_descr.description = ''

            # find repetition directories and ids
            repetition_directories = glob(os.path.join(exp_directory, repetition_directory_template))
            experiment_descr.repetition_directories = repetition_directories
            if experiment_descr.repetition_directories:
                experiment_descr.repetition_directories.sort()

            experiment_descr.repetition_ids = []
            for rep_directory in np.sort(repetition_directories):
                rep_id = re.findall(r'\d+', os.path.basename(rep_directory))[0]
                experiment_descr.repetition_ids.append(int(rep_id))
            experiment_descr.repetition_ids.sort()

            experiment_descriptions[exp_id] = experiment_descr

    return experiment_descriptions


def load_experiment_data(experiment_descriptions: Optional[AttrDict]=None,
                         experiments_directory: Optional[str]=None,
                         allowed_experiments_id_list: Optional[list]=None,
                         denied_experiments_id_list: Optional[list]=None,
                         data_directory: Optional[str]=None,
                         is_load_repetition_data: bool=True,
                         pre_allowed_data_filter: Optional[list]=None,
                         pre_denied_data_filter: Optional[list]=None,
                         post_allowed_data_filter: Optional[list]=None,
                         post_denied_data_filter: Optional[list]=None,
                         on_experiment_data_loaded: Optional[list]=None,
                         on_repetition_data_loaded: Optional[list]=None,
                         allow_pickle: bool = True) -> tuple[AttrDict, AttrDict]:
    """
    Loads logged data from experiments and their repetitions in form of nested dictionaries and numpy arrays.

    [//]: # (TODO: give an example of a file structure and how it is loaded)

    Parameters:
        experiment_descriptions (AttrDict):
            Predefined descriptions of the experiments.
            The descriptions contain the paths to all experiments and their repetitions that should be loaded.
            See [`load_experiment_descriptions`][exputils.data.loading.load_experiment_descriptions] for details.
            Can not be set together with experiments_directory.
        experiments_directory (str):
            Path to the experiments directory.
            Defaults to `'..\experiments'`.
            Can not be set together with experiment_descriptions.
        allowed_experiments_id_list (list):
            List of allowed experiment IDs.
            Only these will be loaded.
            Can not be set together with denied_experiments_id_list.
        denied_experiments_id_list (list):
            List of denied experiment IDs.
            All experiments besides these will be loaded.
            Can not be set together with allowed_experiments_id_list.
        data_directory (str):
            Relative path of the data directories under the experiments and repetitions.
            Defaults to `'./data'`.
        is_load_repetition_data (bool):
            Flag to indicate if repetition data should be loaded.
            Defaults to `True`.
        pre_allowed_data_filter (list):
            List of datasources that will be loaded before the loading callback functions
            (see `on_experiment_data_loaded` and `on_repetition_data_loaded`) are called.
            Thus this data will be given to the callback functions.
            If defined then only these datasources will be loaded.
            The list contains strings with names of datasources.
        pre_denied_data_filter (list):
            List of datasources that will NOT be loaded before the loading callback functions
            (see `on_experiment_data_loaded` and `on_repetition_data_loaded`) are called.
            Thus this data will NOT be given to the callback functions.
            If defined then all existing datasources besides the ones specified will be loaded.
            The list contains strings with names of datasources.
        post_allowed_data_filter (list):
            List of datasources that will be added to the data dictionary that is returned after
            loading and the loading callback functions (see `on_experiment_data_loaded` and `on_repetition_data_loaded`)
            are called.
            If defined then only these datasources will be returned.
            The list contains strings with names of datasources.
        post_denied_data_filter (list):
            List of datasources that will NOT be added to the data dictionary that is returned after
            loading and the loading callback functions (see `on_experiment_data_loaded` and `on_repetition_data_loaded`)
            are called.
            If defined then all existing datasources besides the ones specified will be returned.
            The list contains strings with names of datasources.
        on_experiment_data_loaded (list):
            List of callback functions executed when experiment data is loaded.
            Can be used to modify the data by changing or adding elements.
            Form of the functions: func(exp_id: int, exp_data: AttrDict) -> AttrDict.
        on_repetition_data_loaded (list):
            List of callback functions executed when repetition data is loaded.
            Can be used to modify the data by changing or adding elements.
            Form of the functions: func(exp_id: int, exp_data: AttrDict) -> AttrDict.
        allow_pickle (bool):
            Indicates if loading of pickled objects is allowed.
            Defaults to True. <br>
            :warning: This could allow arbitrary code execution. Only load files you trust!

    Returns:
        data (AttrDict):
            Loaded data.
        experiment_descriptions (AttrDict):
            Experiment descriptions of the loaded data.
            See [`load_experiment_descriptions`][exputils.data.loading.load_experiment_descriptions] for details.
    """

    if experiments_directory is not None and experiment_descriptions is not None:
        raise ValueError('Can not set experiment_directory and experiment_descriptions at the same time!')

    if experiment_descriptions is not None and (allowed_experiments_id_list is not None or denied_experiments_id_list is not None):
        raise ValueError('experiment_descriptions and (allowed_experiments_id_list or denied_experiments_id_list) can not be set at the same time!')

    if allowed_experiments_id_list is not None and denied_experiments_id_list is not None:
        raise ValueError('allowed_experiments_id_list and denied_experiments_id_list can not be set at the same time!')

    if experiment_descriptions is None:
        experiment_descriptions = load_experiment_descriptions(
            experiments_directory=experiments_directory,
            allowed_experiments_id_list=allowed_experiments_id_list,
            denied_experiments_id_list=denied_experiments_id_list
        )
    else:
        experiment_descriptions = experiment_descriptions

    if on_experiment_data_loaded is None:
        on_experiment_data_loaded = []

    if on_repetition_data_loaded is None:
        on_repetition_data_loaded = []

    # load experiments according to the order in the experiment_descriptions
    sorted_experiment_ids = eu.data.get_ordered_experiment_ids_from_descriptions(experiment_descriptions)

    data = collections.OrderedDict()
    for exp_id in sorted_experiment_ids:
        exp_descr = experiment_descriptions[exp_id]

        if 'is_load_data' not in exp_descr or exp_descr['is_load_data']:
            try:
                data[exp_id] = load_single_experiment_data(
                    exp_descr['directory'],
                    data_directory=data_directory,
                    allowed_data_filter=pre_allowed_data_filter,
                    denied_data_filter=pre_denied_data_filter,
                    allow_pickle=allow_pickle)

                for callback_function in on_experiment_data_loaded:
                    callback_function(exp_id, data[exp_id])

                _filter_data(data[exp_id], post_allowed_data_filter, post_denied_data_filter)

            except FileNotFoundError:
                if not exp_descr.repetition_ids or not is_load_repetition_data:
                    warnings.warn('Could find data for experiment {!r} ({!r}). Skipped ...'.format(exp_id, exp_descr['directory']))

            except Exception as e:
                raise Exception('Exception during loading of data for experiment {!r} ({!r})!'.format(exp_id, exp_descr['directory'])) from e

            # load data of each repetition
            if is_load_repetition_data:
                if eu.REPETITION_DATA_KEY in data:
                    warnings.warn('A statistic called {!r} was loaded for experiment data. Can not store repetition data under the same data source name. Skip to load repetition data. Please rename this statistic.'.format(eu.REPETITION_DATA_KEY))
                else:
                    cur_rep_statistics_dict = dict()
                    for rep_id in exp_descr.repetition_ids:
                        cur_rep_directory = os.path.join(exp_descr['directory'], eu.REPETITION_DIRECTORY_TEMPLATE.format(rep_id))
                        try:
                            cur_rep_statistics_dict[rep_id] = load_single_experiment_data(
                                cur_rep_directory,
                                data_directory=data_directory,
                                allowed_data_filter=pre_allowed_data_filter,
                                denied_data_filter=pre_denied_data_filter,
                                allow_pickle=allow_pickle)

                            for callback_function in on_repetition_data_loaded:
                                callback_function(exp_id, rep_id, cur_rep_statistics_dict[rep_id])

                            _filter_data(cur_rep_statistics_dict[rep_id], post_allowed_data_filter, post_denied_data_filter)

                        except FileNotFoundError:
                            warnings.warn('Could not find data for repetition {} of experiment {!r} ({!r}). Skipped ...'.format(rep_id, exp_id, exp_descr['directory']))

                        except Exception as e:
                            raise Exception('Exception during loading of data for repetition {} of experiment {!r} ({!r})!'.format(rep_id, exp_id, exp_descr['directory'])) from e

                    if cur_rep_statistics_dict:
                        # in case no experimental level data exists
                        if exp_id not in data:
                            data[exp_id] = AttrDict()

                        data[exp_id][eu.REPETITION_DATA_KEY] = cur_rep_statistics_dict

    return data, experiment_descriptions


def load_single_experiment_data(experiment_directory: str,
                                data_directory: Optional[str] = None,
                                allowed_data_filter: Optional[list] = None,
                                denied_data_filter: Optional[list] = None,
                                allow_pickle: bool = True) -> AttrDict:
    """
    Loads data for a single experiment which includes all its repetition data.

    Parameters:
        experiment_directory (str):
            Path to the experiment directory.
        data_directory (str):
            Relative path of the data directories under the experiments and repetitions.
            Defaults to `'./data'`.
        allowed_data_filter (list):
            List of datasource names (strings) that will be loaded.
            If defined then only these datasources will be loaded.
        denied_data_filter (list):
            List of datasource names (strings) that will NOT be loaded.
            If defined then all datasources besides the specified ones will be loaded.
        allow_pickle (bool):
            Indicates if loading of pickled objects is allowed.
            Defaults to True. <br>
            :warning: This could allow arbitrary code execution. Only load files you trust!

    Returns:
        data (AttrDict):
            A dictionary containing the loaded data.
    """

    if data_directory is None:
        data_directory = eu.DEFAULT_DATA_DIRECTORY

    # need to allow also logging to be able to load data that is in logging.npz files
    if allowed_data_filter is not None:
        allowed_data_filter.append('logging')

    data = eu.io.load_numpy_files(
        os.path.join(experiment_directory, data_directory),
        allowed_data_filter=allowed_data_filter,
        denied_data_filter=denied_data_filter,
        allow_pickle=allow_pickle)

    # TODO: Refactor - make loading of npz files without the 'logging' sub-directory as a general cases
    if 'logging' in data:
        data.update(data['logging'])
        del data['logging']

    return data


def _filter_data(data, allowed_data_list, denied_data_list):
    # get the data_elements that should be deleted
    delete_keys = [k for k in data.keys() if not eu.misc.is_allowed(k, allowed_list=allowed_data_list, denied_list=denied_data_list)]
    for delete_key in delete_keys:
        del data[delete_key]


def load_experiment_data_single_object(name: str,
                                       experiment_id: Optional[int] = None,
                                       repetition_id: Optional[int] = None,
                                       experiments_directory: Optional[str] = None,
                                       data_directory: Optional[str] = None,
                                       experiment_directory_template: Optional[str] = None,
                                       repetition_directory_template: Optional[str] = None,
                                       add_execution_directory_to_sys_path: bool = True) -> object:
    """
    Loads single object that was logged via the [add_single_object][exputils.data.logging.add_single_object].
    function and saved as a dill file. The file that is either located under the experiments, or a
    single experiment or repetition directory.

    :warning: This could allow arbitrary code execution. Only load files you trust!

    Example:
        ```python
        loaded_obj = load_experiment_data_single_object(
            'my_object',  # name of the object
            experiment_id=100,
            repetition_id=1
        )
        print(loaded_obj)
        ```

    Parameters:
        name (str):
            The name of the object which is the name of the dill file with or without extension.
        experiment_id (int):
            An optional identifier for a specific experiment.
            If not provided, then the object is loaded from the experiments directory.
        repetition_id (int):
            An optional identifier for a specific repetition of the experiment.
            If not provided, then the object is loaded from the experiments or experiment directory.
        experiments_directory (str):
            The root directory where all experiments are stored.
            Defaults to `..\experiments`.
        data_directory (str):
            Relative path of the data directories under the experiments and repetitions.
            Defaults to `'./data'`.
        experiment_directory_template (str):
            Name template of experiment directories.
            Defaults to `'experiment_{:06d}'`.
        repetition_directory_template (str): Template for constructing the repetition directory. Defaults to None.
            Name template of repetition directories.
            Defaults to `'repetition_{:06d}'`.
        add_execution_directory_to_sys_path (bool):
            Whether to add the execution directory to the system path temporailly while loading the object.
            This can be necessary if the object has relational import statements.
            By adding the directory where the object is located temporailly to the python path,
            these import statments can be processed correctly.
            Defaults to True.

    Returns:
        object: The loaded experiment data object.
    """


    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    full_execution_dir_path = experiments_directory

    # only add experiment subfolder if needed
    if experiment_id is not None:

        if experiment_directory_template is None:
            experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE

        experiment_directory = experiment_directory_template.format(experiment_id)

        full_execution_dir_path = os.path.join(full_execution_dir_path, experiment_directory)

        # only add repetition subfolder if needed
        if repetition_id is not None:

            if repetition_directory_template is None:
                repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE

            repetition_directory = repetition_directory_template.format(repetition_id)

            full_execution_dir_path = os.path.join(full_execution_dir_path, repetition_directory)

    if data_directory is None:
        data_directory = eu.DEFAULT_DATA_DIRECTORY

    # construct the full path to the module
    full_dill_path = os.path.join(full_execution_dir_path, data_directory, name)

    # add the directory in which the code was executed to system path
    if add_execution_directory_to_sys_path:
        sys.path.append(full_execution_dir_path)

    obj = eu.io.dill.load_dill(full_dill_path)

    if add_execution_directory_to_sys_path:
        sys.path.pop()

    return obj


def load_experiment_python_module(module_path: str,
                                  experiment_id: Optional[int] = None,
                                  repetition_id: Optional[int] = None,
                                  experiments_directory: Optional[str] = None,
                                  exec_module: bool = True,
                                  experiment_directory_template: Optional[str] = None,
                                  repetition_directory_template: Optional[str] = None,
                                  add_execution_directory_to_sys_path: bool = True) -> ModuleType:
    """
    Loads a Python module dynamically that is either located under the experiments, or a single
    experiment or repetition directory.
    This can be used to load for example the configuration file of a repetition.

    Example:
        ```python
        # load the configuration file of a repetition and print its config dictionary
        config_module = load_experiment_python_module(
            'repetition_config.py',  # name of the configuration file
            experiment_id=100,
            repetition_id=3
        )
        print(config_module.config)
        ```

    :warning: This could allow arbitrary code execution. Only load files you trust!

    Parameters:
        module_path (str):
            The realtive path to the python module file either under the experiments, experiment, or
            repetition directory.
            Which level depends on if an experiment_id and a repetition_id are provided or not.
        experiment_id (int):
            An optional identifier for a specific experiment.
            If not provided, then the module is loaded from the experiments directory.
        repetition_id (int):
            An optional identifier for a specific repetition of the experiment.
            If not provided, then the module is loaded from the experiments or experiment directory.
        experiments_directory (str):
            The root directory where all experiments are stored.
            Defaults to `..\experiments`.
        exec_module (bool):
            If True, the module will be executed after being loaded which means it will be imported.
            Defaults to True.
        experiment_directory_template (str):
            Name template of experiment directories.
            Defaults to `'experiment_{:06d}'`.
        repetition_directory_template (str): Template for constructing the repetition directory. Defaults to None.
            Name template of repetition directories.
            Defaults to `'repetition_{:06d}'`.
        add_execution_directory_to_sys_path (bool):
            If True, the script's execution directory will be added to sys.path.
            Defaults to True.

    Returns:
        module (ModuleType): The loaded Python module object.

    """

    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    full_module_path = experiments_directory

    # only add experiment subfolder if needed
    if experiment_id is not None:

        if experiment_directory_template is None:
            experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE

        experiment_directory = experiment_directory_template.format(experiment_id)

        full_module_path = os.path.join(full_module_path, experiment_directory)

        # only add repetition subfolder if needed
        if repetition_id is not None:

            if repetition_directory_template is None:
                repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE

            repetition_directory = repetition_directory_template.format(repetition_id)

            full_module_path = os.path.join(full_module_path, repetition_directory)

    # construct the full path to the module
    full_module_path = os.path.join(full_module_path, module_path)

    filename = os.path.basename(module_path)
    module_name = filename.replace('.py', '')

    spec = importlib.util.spec_from_file_location(module_name, full_module_path)

    # creates a new module based on spec
    module = importlib.util.module_from_spec(spec)

    if exec_module:
        # add the directory in which the code was executed to system path
        if add_execution_directory_to_sys_path:
            sys.path.append(os.path.dirname(full_module_path))

        spec.loader.exec_module(module)

        if add_execution_directory_to_sys_path:
            sys.path.pop()

    return module



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022, 2023
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from typing import Optional
import exputils as eu
import numpy as np
import os
from glob import glob
from exputils.misc.attrdict import AttrDict


def save_dict_to_numpy_files(data: dict,
                             path: Optional[str] = '.',
                             mode: Optional[str] = 'npy'):
    """Saves a dictionary with numpy arrays to numpy files (either .npy, .npz, or .npz compressed formats).

    Parameters:
        data (dict):
            Dictionary containing the data to be saved, with keys as filenames and values as data to be saved.
        path (str):
            Directory or file path where the numpy files will be saved.
            Default is the current directory.
        mode (str):
            Mode in which to save the data.
            Can be 'npy', 'npz', or 'cnpz'.
            Default is 'npy'.

    Raises:
        ValueError: If an invalid mode is provided.
    """

    # save logs in numpy format if they exist
    if mode.lower() == 'npy':
        eu.io.makedirs(path)
        for name, values in data.items():
            np.save(os.path.join(path, name), values)

    elif mode.lower() == 'npz':
        eu.io.makedirs_for_file(path)
        np.savez(path, **data)

    elif mode.lower() == 'cnpz':
        eu.io.makedirs_for_file(path)
        np.savez_compressed(path, **data)

    else:
        raise ValueError('Unknown numpy logging mode {!r}! Only \'npy\', \'npz\' and \'cnpz\' are allowed.'.format(mode))


def load_numpy_files(directory: str,
                    allowed_data_filter: Optional[list] = None,
                    denied_data_filter: Optional[list] = None,
                    allow_pickle: bool = True) -> AttrDict:
    """Loads numpy files from a specified directory into an AttrDict.

    Parameters:
        directory (str):
            The path to the directory containing the numpy files.
        allowed_data_filter (list, optional):
            A list of allowed file names to be loaded.
            If specified, only files with names in this list will be loaded.
        denied_data_filter (list, optional):
            A list of denied file names to be excluded from loading.
            If specified, files with names in this list will not be loaded.
        allow_pickle (bool):
            Whether to allow loading pickled (serialized) objects.
            Default is True. <br>
            :warning: This could allow arbitrary code execution. Only load files you trust!

    Raises:
        ValueError:
            If both allowed_data_filter and denied_data_filter are specified.
        FileNotFoundError:
            If the specified directory does not exist.
        Exception:
            If an error occurs during loading of a file.

    Returns:
        data (AttrDict):
            Dictionary with loaded data where the keys are file names without extensions
            and the values are the respective numpy arrays.
    """

    if allowed_data_filter is not None and denied_data_filter is not None:
        raise ValueError('in_data_filter and out_data_filter can not both be set, only one or none!')

    if not os.path.isdir(directory):
        raise FileNotFoundError('Directory {!r} does not exist!'.format(directory))

    data = AttrDict()

    for file in glob(os.path.join(directory, '*.npy')):
        stat_name = os.path.splitext(os.path.basename(file))[0]

        if eu.misc.is_allowed(stat_name, allowed_list=allowed_data_filter, denied_list=denied_data_filter):
            try:
                stat_val = np.load(file, allow_pickle=allow_pickle)
            except FileNotFoundError:
                raise
            except Exception as e:
                raise Exception('Exception during loading of file {!r}!'.format(file)) from e

            if len(stat_val.shape) == 0:
                stat_val = stat_val.dtype.type(stat_val)

            data[stat_name] = stat_val

    for file in glob(os.path.join(directory, '*.npz')):
        stat_name = os.path.splitext(os.path.basename(file))[0]
        if eu.misc.is_allowed(stat_name, allowed_list=allowed_data_filter, denied_list=denied_data_filter):
            try:
                stat_vals = AttrDict(np.load(file, allow_pickle=allow_pickle))
            except FileNotFoundError:
                raise
            except Exception as e:
                raise Exception('Exception during loading of file {!r}!'.format(file)) from e

            # remove data that should not be loaded
            keys = [k for k, v in stat_vals.items() if not eu.misc.is_allowed(k, allowed_list=allowed_data_filter, denied_list=denied_data_filter)]
            for x in keys:
                del stat_vals[x]

            # numpy encapsulates scalars as darrays with an empty shape
            # recover the original type
            for substat_name, substat_val in stat_vals.items():
                if len(substat_val.shape) == 0:
                    stat_vals[substat_name] = substat_val.dtype.type(substat_val)

            data[stat_name] = stat_vals

    return data


